/*
 * Copyright 2010 CodeWeavers, Aric Stewart
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */
#include <stdarg.h>
#include <stdlib.h>
#include "windef.h"
#include "winbase.h"
#include "winuser.h"
#include "osxime_private.h"

#include "wine/debug.h"
WINE_DEFAULT_DEBUG_CHANNEL(mac_ime);

enum FUNCTIONS {
    MACTSMCHECKMODEANDIME,
    MACTSMFIXTEXTSERVICE,
    MACTSMREVERTTOPREVIOUSMODE,
    XTOASCII,
    MACTOASCII,
    SHUTDOWN
};

typedef struct _serverPacket {
    HANDLE      StartEvent;
    HANDLE      EndEvent;
    INT         requestedFunction;
    INT         returnCode;
    TSMHandle   TsmHandle;
    HANDLE      threadHandle;
    DWORD       threadId;

    UINT uVKey;
    UINT keysym;
    UINT keycode;
    LPBYTE lpbKeyState;
    LPDWORD lpdwTransKey;
    WCHAR unichar;
    char theChar;
    unsigned long theCode;
    void* hIMC;
} serverPacket;

serverPacket packet;

static CRITICAL_SECTION imeserver_cs;
static CRITICAL_SECTION_DEBUG critsect_debug =
{
    0, 0, &imeserver_cs,
    { &critsect_debug.ProcessLocksList, &critsect_debug.ProcessLocksList },
      0, 0, { (DWORD_PTR)(__FILE__ ": imeserver_cs") }
};
static CRITICAL_SECTION imeserver_cs = { &critsect_debug, -1, 0, 0, 0, 0 };

static DWORD WINAPI ThreadMain(LPVOID param)
{
    BOOL exitNow = FALSE;
    serverPacket *spacket = (serverPacket*)param;

    MacInstallAppEventHandler(spacket->TsmHandle);
    MacNewTSMDocument(spacket->TsmHandle,(long)0xfeedface);

    SetEvent(spacket->EndEvent);
    while (!exitNow)
    {
        WaitForSingleObject(spacket->StartEvent, INFINITE);

        switch (spacket->requestedFunction)
        {
            case MACTSMCHECKMODEANDIME :
                spacket->returnCode = MacTSMCheckModeAndIME(spacket->TsmHandle);
                break;
            case MACTSMFIXTEXTSERVICE:
                MacTSMFixTextService(spacket->TsmHandle);
                break;
            case MACTSMREVERTTOPREVIOUSMODE:
                MacTSMRevertToPreviousMode(spacket->TsmHandle);
                break;
            case XTOASCII:
                spacket->returnCode = XToAscii(spacket->uVKey, spacket->keysym,
spacket->keycode, spacket->lpbKeyState, spacket->lpdwTransKey, spacket->hIMC);
                break;
            case MACTOASCII:
                spacket->returnCode = MacToAscii(spacket->uVKey, spacket->unichar,
spacket->theChar, spacket->theCode, spacket->lpbKeyState, spacket->lpdwTransKey,
            spacket->hIMC);
                break;
            case SHUTDOWN:
                MacRemoveAppEventHandler(spacket->TsmHandle);
                exitNow = TRUE;
                break;
            default:
                ERR("Unknown function requested\n");
        }
        SetEvent(spacket->EndEvent);
    }
    return 0;
}

TSMHandle Server_Init(void)
{
    TRACE("->\n");
    packet.threadHandle = NULL;
    packet.TsmHandle = MacInitTSMAwareApplication();
    TRACE("<-\n");
    return packet.TsmHandle;
}

void Server_Shutdown(void)
{
    if (packet.threadHandle != NULL)
    {
        packet.requestedFunction = SHUTDOWN;
        SetEvent(packet.StartEvent);
        packet.threadHandle = NULL;
    }
    if (packet.TsmHandle)
        MacCloseTSMAwareApplication(packet.TsmHandle);
    TRACE("Server:End.\n");
}

static void WaitServer(void)
{
    SetEvent(packet.StartEvent);
    WaitForSingleObject(packet.EndEvent, INFINITE);
}

static void CheckServer(void)
{
    if (!packet.threadHandle)
    {
        packet.StartEvent = CreateEventW(NULL, FALSE, FALSE, NULL);
        packet.EndEvent = CreateEventW(NULL, FALSE, FALSE, NULL);
        packet.threadHandle = CreateThread(NULL, 0, ThreadMain, &packet, 0, &packet.threadId);
        WaitForSingleObject(packet.EndEvent, INFINITE);
    }
}

int Server_MacTSMCheckModeAndIME(void)
{
    int rc;
    TRACE("->\n");
    EnterCriticalSection(&imeserver_cs);
    CheckServer();
    packet.requestedFunction = MACTSMCHECKMODEANDIME;
    WaitServer();
    rc = packet.returnCode;
    LeaveCriticalSection(&imeserver_cs);
    TRACE("<-\n");
    return rc;
}

void Server_MacTSMFixTextService(void)
{
    TRACE("->\n");
    EnterCriticalSection(&imeserver_cs);
    CheckServer();
    packet.requestedFunction = MACTSMFIXTEXTSERVICE;
    WaitServer();
    LeaveCriticalSection(&imeserver_cs);
    TRACE("<-\n");
}

void Server_MacTSMRevertToPreviousMode(void)
{
    TRACE("->\n");
    EnterCriticalSection(&imeserver_cs);
    CheckServer();
    packet.requestedFunction = MACTSMREVERTTOPREVIOUSMODE;
    WaitServer();
    LeaveCriticalSection(&imeserver_cs);
    TRACE("<-\n");
}


int Server_XToAscii(UINT uVKey, UINT keysym, UINT keycode, CONST LPBYTE lpbKeyState,
    LPDWORD lpdwTransKey, void* hIMC)
{
    int rc;
    TRACE("->\n");
    EnterCriticalSection(&imeserver_cs);
    CheckServer();
    packet.requestedFunction = XTOASCII;
    packet.uVKey = uVKey;
    packet.keysym = keysym;
    packet.keycode = keycode;
    packet.lpbKeyState = lpbKeyState;
    packet.lpdwTransKey = lpdwTransKey;
    packet.hIMC = hIMC;
    WaitServer();
    rc = packet.returnCode;
    LeaveCriticalSection(&imeserver_cs);
    TRACE("<-\n");
    return rc;
}

int Server_MacToAscii(UINT uVKey, WCHAR unichar, char theChar, unsigned long theCode, CONST LPBYTE lpbKeyState, LPDWORD lpdwTransKey, void* hIMC)
{
    int rc;
    TRACE("->\n");
    EnterCriticalSection(&imeserver_cs);
    CheckServer();
    packet.requestedFunction = MACTOASCII;
    packet.uVKey = uVKey;
    packet.unichar = unichar;
    packet.theChar = theChar;
    packet.theCode = theCode;
    packet.lpbKeyState = lpbKeyState;
    packet.lpdwTransKey = lpdwTransKey;
    packet.hIMC = hIMC;
    WaitServer();
    rc = packet.returnCode;
    LeaveCriticalSection(&imeserver_cs);
    TRACE("<-\n");
    return rc;
}
